import os
import re
import time
import logging
import collections
from itertools import chain

from .vendor.Qt import QtWidgets, QtCore, QtGui, QtCompat
from .vendor import qargparse, QtImageViewer

from . import resources as res, model, delegates, util
from . import allzparkconfig

try:
    from localz import lib as localz
except ImportError:
    localz = None

px = res.px


class AbstractDockWidget(QtWidgets.QDockWidget):
    """Default HTML <b>docs</b>"""

    icon = ""
    advanced = False

    message = QtCore.Signal(str)  # Handled by main window

    def __init__(self, title, parent=None):
        super(AbstractDockWidget, self).__init__(title, parent)
        self.layout().setContentsMargins(px(15), px(15), px(15), px(15))

        panels = {
            "help": QtWidgets.QLabel(),
            "body": QtWidgets.QStackedWidget(),
        }

        for name, widget in panels.items():
            widget.setAttribute(QtCore.Qt.WA_StyledBackground)
            widget.setObjectName("dock%s" % name.title())

        central = QtWidgets.QWidget()

        layout = QtWidgets.QVBoxLayout(central)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(px(5))
        layout.addWidget(panels["help"])
        layout.addWidget(panels["body"])

        if self.__doc__:
            panels["help"].setText(self.__doc__.splitlines()[0])
        else:
            panels["help"].hide()

        self.__panels = panels

        QtWidgets.QDockWidget.setWidget(self, central)

    def setWidget(self, widget):
        body = self.__panels["body"]

        while body.widget(0):
            body.removeWidget(body.widget(0))

        body.addWidget(widget)


class App(AbstractDockWidget):
    """Aggregated information about the currently selected profile"""

    icon = "Alert_Info_32"

    def __init__(self, ctrl, parent=None):
        super(App, self).__init__("App", parent)
        self.setAttribute(QtCore.Qt.WA_StyledBackground)
        self.setObjectName("App")

        panels = {
            "central": QtWidgets.QWidget(),
            "shortcuts": QtWidgets.QWidget(),
            "footer": QtWidgets.QWidget(),
        }

        widgets = {
            "icon": QtWidgets.QLabel(),
            "label": QtWidgets.QLabel("Autodesk Maya"),
            "version": QtWidgets.QComboBox(),
            "tool": QtWidgets.QToolButton(),

            "commands": SlimTableView(),
            "extras": SlimTableView(),
            "lastUsed": QtWidgets.QLabel(),

            "args": qargparse.QArgumentParser([
                qargparse.Choice("tool", help=(
                    "Which executable within the context of this application"
                )),
            ]),

            # Shortcuts
            "environment": QtWidgets.QToolButton(),
            "packages": QtWidgets.QToolButton(),
            "terminal": QtWidgets.QToolButton(),

            "launchBtn": QtWidgets.QPushButton("Launch"),
        }

        # Expose to CSS
        for name, widget in chain(panels.items(), widgets.items()):
            widget.setAttribute(QtCore.Qt.WA_StyledBackground)
            widget.setObjectName(name)

        layout = QtWidgets.QHBoxLayout(panels["shortcuts"])
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(widgets["environment"])
        layout.addWidget(widgets["packages"])
        layout.addWidget(widgets["terminal"])
        layout.addWidget(QtWidgets.QWidget(), 1)  # push to the left

        layout = QtWidgets.QHBoxLayout(panels["footer"])
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(widgets["launchBtn"])

        layout = QtWidgets.QGridLayout(panels["central"])
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setHorizontalSpacing(px(10))
        layout.setVerticalSpacing(0)

        def Spacer():
            return QtWidgets.QLabel("")

        layout.addWidget(widgets["icon"], 0, 0, 2, 1)
        layout.addWidget(widgets["label"], 0, 1, QtCore.Qt.AlignTop)
        layout.addWidget(widgets["lastUsed"], 1, 1, QtCore.Qt.AlignTop)
        layout.addWidget(widgets["args"], 5, 0, 1, 2)
        layout.addWidget(widgets["commands"], 6, 0, 1, 2)
        layout.addWidget(QtWidgets.QWidget(), 15, 0)
        layout.setColumnStretch(1, 1)
        layout.setRowStretch(15, 1)
        layout.addWidget(panels["footer"], 40, 0, 1, 2)

        widgets["icon"].setPixmap(res.pixmap("Alert_Info_32"))
        widgets["environment"].setIcon(res.icon(Environment.icon))
        widgets["packages"].setIcon(res.icon(Packages.icon))
        widgets["terminal"].setIcon(res.icon(Console.icon))

        widgets["args"].changed.connect(self.on_arg_changed)

        widgets["launchBtn"].setCheckable(True)
        widgets["launchBtn"].clicked.connect(self.on_launch_clicked)

        proxy_model = model.ProxyModel(ctrl.models["commands"])
        widgets["commands"].setModel(proxy_model)

        self._ctrl = ctrl
        self._panels = panels
        self._widgets = widgets
        self._proxy = proxy_model

        self.setWidget(panels["central"])

    def on_launch_clicked(self):
        self._ctrl.launch()

    def on_arg_changed(self, arg):
        if arg["name"] not in ("tool",):
            return

        ctrl = self._ctrl
        model = ctrl.models["apps"]
        app_name = ctrl.state["appRequest"]
        app_index = model.findIndex(app_name)
        value = arg.read()
        model.setData(app_index, value, arg["name"])

        if arg["name"] == "tool":
            ctrl.select_tool(value)

    def refresh(self, index):
        name = index.data(QtCore.Qt.DisplayRole)
        icon = index.data(QtCore.Qt.DecorationRole)

        if icon:
            icon = icon.pixmap(QtCore.QSize(px(32), px(32)))
            self._widgets["icon"].setPixmap(icon)

        self._widgets["label"].setText(name)

        last_used = self._ctrl.state.retrieve("app/%s/lastUsed" % name)
        last_used = time.strftime(
            "%Y-%m-%d %H:%M:%S", time.localtime(float(last_used))
        ) if last_used else "Never"
        last_used = "Last used: %s" % last_used

        self._widgets["lastUsed"].setText("%s" % last_used)

        model = index.model()
        tools = model.data(index, "tools")
        default_tool = model.data(index, "tool") or tools[0]
        arg = self._widgets["args"].find("tool")
        arg.reset(tools[:], default_tool)

        self._proxy.setup(include=[
            ("appRequest", name),
            ("running", "running"),
        ])


class Console(AbstractDockWidget):
    """Debugging information, mostly for developers"""

    icon = "Prefs_Screen_32"

    def __init__(self, parent=None):
        super(Console, self).__init__("Console", parent)
        self.setAttribute(QtCore.Qt.WA_StyledBackground)
        self.setObjectName("Console")

        panels = {
            "central": QtWidgets.QWidget()
        }

        widgets = {
            "text": QtWidgets.QTextEdit()
        }

        self.setWidget(panels["central"])

        widgets["text"].setReadOnly(True)

        layout = QtWidgets.QVBoxLayout(panels["central"])
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(widgets["text"])

        self._widgets = widgets

    def append(self, line, level=logging.INFO):
        color = {
            logging.DEBUG: "<font color=\"grey\">",
            logging.WARNING: "<font color=\"darkorange\">",
            logging.ERROR: "<font color=\"red\">",
            logging.CRITICAL: "<font color=\"red\">",
        }.get(level, "<font color=\"#222\">")

        line = "%s%s</font><br>" % (color, line)

        cursor = self._widgets["text"].textCursor()
        cursor.movePosition(QtGui.QTextCursor.End)

        self._widgets["text"].setTextCursor(cursor)
        self._widgets["text"].insertHtml(line)

        scrollbar = self._widgets["text"].verticalScrollBar()
        scrollbar.setValue(scrollbar.maximum())


class Packages(AbstractDockWidget):
    """Packages associated with the currently selected application"""

    icon = "File_Archive_32"
    advanced = True

    def __init__(self, ctrl, parent=None):
        super(Packages, self).__init__("Packages", parent)
        self.setAttribute(QtCore.Qt.WA_StyledBackground)
        self.setObjectName("Packages")

        panels = {
            "central": QtWidgets.QTabWidget(),
        }

        pages = {
            "packages": QtWidgets.QWidget(),
        }

        args = [
            qargparse.Boolean(
                "useDevelopmentPackages",
                default=ctrl._state.retrieve("useDevelopmentPackages"),
                help="Include development packages in the resolve"),
            qargparse.String(
                "patch",
                default=ctrl._state.retrieve("patch"),
                help="Add packages to context"),
        ]

        if localz:
            args.insert(0, qargparse.Boolean(
                "useLocalizedPackages",
                default=ctrl._state.retrieve("useLocalizedPackages", True),
                help="Include localised packages in the resolve"),
            )

        widgets = {
            "args": qargparse.QArgumentParser(args),
            "view": SlimTableView(),
            "status": QtWidgets.QStatusBar(),
        }

        self.setWidget(panels["central"])

        panels["central"].addTab(pages["packages"], "Packages")

        layout = QtWidgets.QVBoxLayout(pages["packages"])
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(widgets["args"])
        layout.addWidget(widgets["view"], 1)
        layout.addWidget(widgets["status"])

        widgets["view"].setStretch(4)
        widgets["view"].setItemDelegate(delegates.Package(ctrl, self))
        widgets["view"].setEditTriggers(widgets["view"].DoubleClicked)
        widgets["view"].verticalHeader().setDefaultSectionSize(px(20))
        widgets["view"].customContextMenuRequested.connect(self.on_right_click)

        widgets["status"].setSizeGripEnabled(False)

        widgets["args"].changed.connect(self.on_argument_changed)
        ctrl.resetted.connect(self.on_resetted)

        self._ctrl = ctrl
        self._panels = panels
        self._pages = pages
        self._widgets = widgets

    def on_argument_changed(self, arg):
        if arg["name"] == "useDevelopmentPackages":
            self._ctrl._state.store("useDevelopmentPackages", arg.read())
            self._ctrl.reset()

        if arg["name"] == "useLocalizedPackages":
            self._ctrl._state.store("useLocalizedPackages", arg.read())
            self._ctrl.reset()

        if arg["name"] == "patch":
            self._ctrl._state.store("patch", arg.read())
            self._ctrl.reset()

    def on_resetted(self):
        patch = self._ctrl.state.retrieve("patch", "")
        arg = self._widgets["args"].find("patch")
        arg._write(patch)
        arg._previous = patch

    def set_model(self, model_):
        proxy_model = model.ProxyModel(model_)
        self._widgets["view"].setModel(proxy_model)

        model_.modelReset.connect(self.on_model_changed)
        model_.dataChanged.connect(self.on_model_changed)

    def on_model_changed(self):
        model = self._widgets["view"].model()
        model = model.sourceModel()

        package_count = model.rowCount()
        override_count = len([i for i in model.items if i["override"]])
        disabled_count = len([i for i in model.items if i["disabled"]])

        self._widgets["status"].showMessage(
            "%d Packages, %d Overridden, %d Disabled" % (
                package_count,
                override_count,
                disabled_count,
            ))

    def on_right_click(self, position):
        view = self._widgets["view"]
        index = view.indexAt(position)

        if not index.isValid():
            # Clicked outside any item
            return

        model_ = index.model()
        menu = MenuWithTooltip(self)
        edit = QtWidgets.QAction("Edit", menu)
        disable = QtWidgets.QAction("Disable", menu)
        default = QtWidgets.QAction("Set to default", menu)
        earliest = QtWidgets.QAction("Set to earliest", menu)
        latest = QtWidgets.QAction("Set to latest", menu)
        openfile = QtWidgets.QAction("Open file location", menu)
        copyfile = QtWidgets.QAction("Copy file location", menu)
        localize = QtWidgets.QAction("Localise selected...", menu)
        localize_related = QtWidgets.QAction("Localise related...", menu)
        localize_all = QtWidgets.QAction("Localise all...", menu)
        delocalize = QtWidgets.QAction("Delocalise selected...", menu)

        disable.setCheckable(True)
        disable.setChecked(model_.data(index, "disabled"))

        menu.addAction(edit)
        menu.addSeparator()
        menu.addAction(default)
        menu.addAction(earliest)
        menu.addAction(latest)
        menu.addSeparator()
        menu.addAction(openfile)
        menu.addAction(copyfile)

        if localz:
            enabled = True
            tooltip = None

            if index.data(model.LocalizingRole):
                enabled = False
                tooltip = "Localisation in progress..."

            elif model_.data(index, "state") in ("(dev)", "(localised)"):
                tooltip = "Package already local"
                enabled = False

            elif not model_.data(index, "relocatable"):
                tooltip = "Package does not support localisation"
                enabled = False

            menu.addSeparator()
            for action in (localize,
                           localize_related,
                           localize_all,
                           ):
                menu.addAction(action)
                action.setToolTip(tooltip or "")
                action.setEnabled(enabled)

            menu.addSeparator()
            menu.addAction(delocalize)
            delocalize.setEnabled(
                model_.data(index, "state") == "(localised)"
            )

            # Not yet implemented
            localize_all.setEnabled(False)
            localize_related.setEnabled(False)

        def on_edit():
            self._widgets["view"].edit(index)

        def on_default():
            package = model_.data(index, "package")
            self._ctrl.patch(package.name)
            self.message.emit("Package set to default")

        def on_earliest():
            versions = model_.data(index, "versions")
            earliest = versions[0]
            package = model_.data(index, "package")
            self._ctrl.patch("%s==%s" % (package.name, earliest))

            self.message.emit("Package set to earliest")

        def on_latest():
            versions = model_.data(index, "versions")
            latest = versions[-1]
            package = model_.data(index, "package")
            self._ctrl.patch("%s==%s" % (package.name, latest))

            self.message.emit("Package set to latest version")

        def on_openfile():
            package = model_.data(index, "package")
            fname = os.path.join(package.root, "package.py")
            util.open_file_location(fname)
            self.message.emit("Opened %s" % fname)

        def on_copyfile():
            package = model_.data(index, "package")
            fname = os.path.join(package.root, "package.py")
            clipboard = QtWidgets.QApplication.instance().clipboard()
            clipboard.setText(fname)
            self.message.emit("Copied %s" % fname)

        def on_disable():
            model_.setData(index, None, "override")
            model_.setData(index, disable.isChecked(), "disabled")
            self.message.emit("Package disabled")

        def on_localize():
            name = model_.data(index, "name")
            self._ctrl.localize(name)
            model_.setData(index, "(localising..)", "state")
            model_.setData(index, True, model.LocalizingRole)

        def on_delocalize():
            name = model_.data(index, "name")
            self._ctrl.delocalize(name)
            model_.setData(index, "(delocalising..)", "state")
            model_.setData(index, True, model.LocalizingRole)

        edit.triggered.connect(on_edit)
        disable.triggered.connect(on_disable)
        default.triggered.connect(on_default)
        earliest.triggered.connect(on_earliest)
        latest.triggered.connect(on_latest)
        openfile.triggered.connect(on_openfile)
        copyfile.triggered.connect(on_copyfile)
        localize.triggered.connect(on_localize)
        delocalize.triggered.connect(on_delocalize)

        menu.move(QtGui.QCursor.pos())
        menu.show()


class Context(AbstractDockWidget):
    """Full context relative the currently selected application"""

    icon = "App_Generic_4_32"
    advanced = True

    def __init__(self, ctrl, parent=None):
        super(Context, self).__init__("Context", parent)
        self.setAttribute(QtCore.Qt.WA_StyledBackground)
        self.setObjectName("Context")

        panels = {
            "central": QtWidgets.QTabWidget(),
        }

        pages = {
            "context": QtWidgets.QWidget(),
            "graph": QtWidgets.QWidget(),
        }

        widgets = {
            "view": QtWidgets.QTreeView(),
            "graph": QtImageViewer.QtImageViewer(),
            "generateGraph": QtWidgets.QPushButton("Update"),
            "graphHotkeys": QtWidgets.QLabel(),
            "overlay": QtWidgets.QWidget(),
        }

        # Expose to CSS
        for name, widget in chain(panels.items(),
                                  pages.items(),
                                  widgets.items()):
            widget.setAttribute(QtCore.Qt.WA_StyledBackground)
            widget.setObjectName(name)

        layout = QtWidgets.QVBoxLayout(pages["context"])
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(widgets["view"])

        layout = QtWidgets.QVBoxLayout(pages["graph"])
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(widgets["graph"], 1)

        layout = QtWidgets.QVBoxLayout(widgets["overlay"])
        layout.addWidget(widgets["graphHotkeys"])
        layout.addWidget(widgets["generateGraph"])
        layout.addWidget(QtWidgets.QWidget(), 1)

        panels["central"].addTab(pages["context"], "Context")
        panels["central"].addTab(pages["graph"], "Graph")

        ctrl.application_changed.connect(self.on_application_changed)

        widgets["overlay"].setParent(pages["graph"])
        widgets["overlay"].show()

        widgets["generateGraph"].clicked.connect(self.on_generate_clicked)
        widgets["graphHotkeys"].setText("""\
            <font color=\"steelblue\"><b>Hotkeys</b></font>
            <br>
            <br>
            - <b>Pan</b>: Left mouse <br>
            - <b>Zoom</b>: Right mouse + drag <br>
            - <b>Reset</b>: Double-click right mouse <br>
        """)

        self._ctrl = ctrl
        self._panels = panels
        self._widgets = widgets
        self._model = None

        widgets["view"].setSortingEnabled(True)
        self.setWidget(panels["central"])

    def set_model(self, model_):
        proxy_model = QtCore.QSortFilterProxyModel()
        proxy_model.setSourceModel(model_)
        self._widgets["view"].setModel(proxy_model)
        self._model = model_

    def on_generate_clicked(self):
        pixmap = self._ctrl.graph()

        if not pixmap:
            self._widgets["graphHotkeys"].setText(
                "<b>GraphViz not found</b>"
                "<br>"
                "<br>"
                "This feature requires `dot` on PATH<br>"
                "See <a href=https://allzpark.com>allzpark.com</a> "
                "for details."
            )
            self._widgets["generateGraph"].hide()
            return

        self._widgets["graph"].setImage(pixmap)
        self._widgets["graph"]._pixmapHandle.setGraphicsEffect(None)

    def on_application_changed(self):
        if not self._widgets["graph"]._pixmapHandle:
            return

        grayscale = QtWidgets.QGraphicsColorizeEffect()
        grayscale.setColor(QtGui.QColor(0, 0, 0))
        self._widgets["graph"]._pixmapHandle.setGraphicsEffect(grayscale)


class Environment(AbstractDockWidget):
    """Full environment relative the currently selected application"""

    icon = "App_Heidi_32"
    advanced = True

    def __init__(self, ctrl, parent=None):
        super(Environment, self).__init__("Environment", parent)
        self.setAttribute(QtCore.Qt.WA_StyledBackground)
        self.setObjectName("Environment")

        panels = {
            "central": QtWidgets.QTabWidget(),
        }

        pages = {
            "environment": QtWidgets.QTreeView(),
            "editor": EnvironmentEditor(),
        }

        widgets = {
        }

        layout = QtWidgets.QVBoxLayout(panels["central"])
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(pages["environment"])

        pages["environment"].setSortingEnabled(True)
        self.setWidget(panels["central"])

        pages["editor"].applied.connect(self.on_env_applied)

        panels["central"].addTab(pages["environment"], "Context")
        panels["central"].addTab(pages["editor"], "User")

        user_env = ctrl.state.retrieve("userEnv", {})
        pages["editor"].from_environment(user_env)
        pages["editor"].warning.connect(self.on_env_warning)

        self._ctrl = ctrl
        self._panels = panels
        self._pages = pages
        self._widgets = widgets

    def set_model(self, model_):
        proxy_model = QtCore.QSortFilterProxyModel()
        proxy_model.setSourceModel(model_)
        self._pages["environment"].setModel(proxy_model)

    def on_env_applied(self, env):
        self._ctrl.state.store("userEnv", env)
        self._ctrl.info("User environment successfully saved")

    def on_env_warning(self, message):
        self._ctrl.warning(message)


class TextEditWithFocus(QtWidgets.QTextEdit):
    focusLost = QtCore.Signal()

    def focusOutEvent(self, event):
        super(TextEditWithFocus, self).focusOutEvent(event)
        self.focusLost.emit()


class Commands(AbstractDockWidget):
    """Currently running commands"""

    icon = "App_Pulse_32"
    advanced = True

    def __init__(self, parent=None):
        super(Commands, self).__init__("Commands", parent)
        self.setAttribute(QtCore.Qt.WA_StyledBackground)
        self.setObjectName("Commands")

        panels = {
            "central": QtWidgets.QWidget(),
            "body": QtWidgets.QWidget(),
            "footer": QtWidgets.QWidget(),
        }

        widgets = {
            "view": SlimTableView(),
        }

        layout = QtWidgets.QVBoxLayout(panels["central"])
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(panels["body"])
        layout.addWidget(panels["footer"])

        layout = QtWidgets.QVBoxLayout(panels["body"])
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(widgets["view"])

        layout = QtWidgets.QVBoxLayout(panels["footer"])
        layout.setContentsMargins(0, 0, 0, 0)

        widgets["view"].setSortingEnabled(True)
        widgets["view"].customContextMenuRequested.connect(self.on_right_click)

        self._panels = panels
        self._widgets = widgets

        self.setWidget(panels["central"])

    def set_model(self, model_):
        proxy_model = model.ProxyModel(model_)
        self._widgets["view"].setModel(proxy_model)

    def on_right_click(self, position):
        view = self._widgets["view"]
        index = view.indexAt(position)

        if not index.isValid():
            # Clicked outside any item
            return

        model = index.model()

        menu = QtWidgets.QMenu(self)
        copy_pid = QtWidgets.QAction("Copy pid", menu)

        def on_copy_pid():
            clipboard = QtWidgets.QApplication.instance().clipboard()
            command = model.data(index, "object")
            pid = str(command.pid)
            clipboard.setText(pid)
            self.message.emit("Copying %s" % pid)

        if model.data(index, "running") != "killed":
            copy_pid.triggered.connect(on_copy_pid)
        else:
            copy_pid.setEnabled(False)

        menu.addAction(copy_pid)
        menu.move(QtGui.QCursor.pos())

        menu.show()


class EnvironmentEditor(QtWidgets.QWidget):
    applied = QtCore.Signal(dict)  # environment
    warning = QtCore.Signal(str)  # message

    def __init__(self, parent=None):
        super(EnvironmentEditor, self).__init__(parent)

        widgets = {
            "textEdit": TextEditWithFocus(),
        }

        font = self.font()
        font.setFamily("Courier")
        font.setFixedPitch(True)
        font.setPointSize(10)

        widgets["textEdit"].setFont(font)

        layout = QtWidgets.QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(px(2))
        layout.addWidget(widgets["textEdit"])

        self._widgets = widgets
        self._edited = False

        widgets["textEdit"].focusLost.connect(self.on_focus_lost)
        widgets["textEdit"].textChanged.connect(self.on_text_changed)
        shortcut = QtWidgets.QShortcut(QtGui.QKeySequence("Ctrl+S"), self)
        shortcut.activated.connect(self.on_apply_clicked)

    def on_apply_clicked(self):
        if not self._edited:
            self.warning.emit("Already saved")
            return

        self.on_focus_lost()

    def from_environment(self, environ):
        if not environ:
            return self._widgets["textEdit"].setPlainText("""\
# Provide your own additional environment variables here
#
# MY_VARIABLE=1
# YOUR_VARIABLE=2
#
# NOTE: Variables are applied *before* the context, which
# means your packages overwrite anything you provide here.
""")

        text = "\n".join([
            "%s=%s" % (key, value)
            for key, value in environ.items()
        ])

        self._widgets["textEdit"].setPlainText(text)

    def to_environment(self):
        """Serialise text to dictionary"""

        # Maintain order as written
        env = collections.OrderedDict()

        textedit = self._widgets["textEdit"]
        for line in textedit.toPlainText().splitlines():
            if line.startswith("#"):
                continue

            try:
                key, value = line.split("=")
            except Exception:
                continue

            key = key.rstrip(" ")  # Tailing space
            value = value.strip(" ")  # Leading space

            # Validate key
            validator = re.compile(r'^[A-Za-z0-9\._]+$')
            if not validator.match(key):
                self.warning.emit("Invalid key: %s" % key)
                continue

            env[key] = value
        return env

    def on_focus_lost(self):
        if not self._edited:
            return

        env = self.to_environment()
        self.applied.emit(env)
        self._edited = False

    def on_text_changed(self):
        # Prevent focus from triggering serialisation,
        # if it had already been serialised with Ctrl+S
        self._edited = True


class CssEditor(QtWidgets.QWidget):
    applied = QtCore.Signal(str)  # css

    def __init__(self, parent=None):
        super(CssEditor, self).__init__(parent)

        widgets = {
            "textEdit": QtWidgets.QTextEdit(),
            "apply": QtWidgets.QPushButton("Apply"),
        }

        # Cannot set via CSS, as we need to query the
        # size of it in order to get the tab width below..
        font = self.font()
        font.setFamily("Courier")
        font.setFixedPitch(True)
        font.setPointSize(10)

        widgets["textEdit"].setFont(font)

        layout = QtWidgets.QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(px(2))
        layout.addWidget(widgets["textEdit"])
        layout.addWidget(widgets["apply"])

        widgets["apply"].clicked.connect(self.on_apply_clicked)

        try:
            width = 2 * widgets["textEdit"].fontMetrics().width(" ")
            widgets["textEdit"].setTabStopWidth(width)
        except AttributeError:
            # Qt 5+ only
            pass

        self._widgets = widgets
        self._highlighter = CssHighlighter(widgets["textEdit"].document())

        shortcut = QtWidgets.QShortcut(QtGui.QKeySequence("Ctrl+S"), self)
        shortcut.activated.connect(self.on_apply_clicked)

    def on_apply_clicked(self):
        textedit = self._widgets["textEdit"]
        self.applied.emit(textedit.toPlainText())


class Preferences(AbstractDockWidget):
    """Preferred settings relative the current user"""

    icon = "Action_GoHome_32"

    options = [
        qargparse.Info("startupProfile", help=(
            "Load this profile on startup"
        )),
        qargparse.Info("startupApplication", help=(
            "Load this application on startup"
        )),

        qargparse.Separator("Theme"),

        qargparse.Info("primaryColor", default="white", help=(
            "Main color of the GUI"
        )),
        qargparse.Info("secondaryColor", default="steelblue", help=(
            "Secondary color of the GUI"
        )),

        qargparse.Button("resetLayout", help=(
            "Reset stored layout to their defaults"
        )),

        qargparse.Separator("Settings"),

        qargparse.Boolean("smallIcons", enabled=False, help=(
            "Draw small icons"
        )),
        qargparse.Boolean("allowMultipleDocks", help=(
            "Allow more than one dock to exist at a time"
        )),
        qargparse.Boolean("showAdvancedControls", help=(
            "Show developer-centric controls"
        )),
        qargparse.Boolean("showAllApps", help=(
            "List everything from allzparkconfig:applications\n"
            "not just the ones specified for a given profile."
        )),
        qargparse.Boolean("showHiddenApps", help=(
            "Show apps with metadata['hidden'] = True"
        )),

        qargparse.Boolean("patchWithFilter", help=(
            "Use the current exclusion filter when patching.\n"
            "This enables patching of packages outside of a filter, \n"
            "such as *.beta packages, with every other package still \n"
            "qualifying for that filter."
        )),
        qargparse.Integer("clearCacheTimeout", minimum=1, default=10, help=(
            "Clear package repository cache at this interval, in seconds. \n\n"

            "Default 10. (Requires restart)\n\n"

            "Normally, filesystem calls like `os.listdir` are cached \n"
            "so as to avoid unnecessary calls. However, whenever a new \n"
            "version of a package is released, it will remain invisible \n"
            "until this cache is cleared. \n\n"

            "Clearing ths cache should have a very small impact on \n"
            "performance and is safe to do frequently. It has no effect \n"
            "on memcached which has a much greater impact on performanc."
        )),

        qargparse.String(
            "exclusionFilter",
            default=allzparkconfig.exclude_filter,
            help="Exclude versions that match this expression"),

        qargparse.Separator("System"),

        # Provided by controller
        qargparse.Info("pythonExe"),
        qargparse.Info("pythonVersion"),
        qargparse.Info("qtVersion"),
        qargparse.Info("qtBinding"),
        qargparse.Info("qtBindingVersion"),
        qargparse.Info("rezLocation"),
        qargparse.Info("rezVersion"),
        qargparse.Info("rezConfigFile"),
        qargparse.Info("memcachedURI"),
        qargparse.InfoList("rezPackagesPath"),
        qargparse.InfoList("rezLocalPath"),
        qargparse.InfoList("rezReleasePath"),
        qargparse.Info("settingsPath"),
    ]

    def __init__(self, window, ctrl, parent=None):
        super(Preferences, self).__init__("Preferences", parent)
        self.setAttribute(QtCore.Qt.WA_StyledBackground)
        self.setObjectName("Preferences")

        panels = {
            "central": QtWidgets.QTabWidget(),
        }

        pages = {
            "settings": QtWidgets.QWidget(),
            "cssEditor": CssEditor(),
        }

        widgets = {
            "options": qargparse.QArgumentParser(
                self.options, storage=ctrl._storage)
        }

        layout = QtWidgets.QVBoxLayout(pages["settings"])
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(widgets["options"])

        scroll = QtWidgets.QScrollArea()
        scroll.setWidget(pages["settings"])
        scroll.setWidgetResizable(True)

        widgets["options"].changed.connect(self.handler)
        pages["cssEditor"].applied.connect(self.on_css_applied)

        panels["central"].addTab(scroll, "Settings")
        panels["central"].addTab(pages["cssEditor"], "CSS")

        self._panels = panels
        self._pages = pages
        self._widgets = widgets
        self._ctrl = ctrl
        self._window = window

        user_css = ctrl.state.retrieve("userCss", "")
        pages["cssEditor"]._widgets["textEdit"].setPlainText(user_css)

        self.setWidget(panels["central"])

    def handler(self, argument):
        self._window.on_setting_changed(argument)

    def on_css_applied(self, css):
        self._ctrl.state.store("userCss", css)
        self._window.setStyleSheet("\n".join([
            self._window._originalcss, css]))
        self._window.tell("Applying css..")


class SlimTableView(QtWidgets.QTableView):
    doSort = QtCore.Signal(int, QtCore.Qt.SortOrder)

    def __init__(self, parent=None):
        super(SlimTableView, self).__init__(parent)
        self.setShowGrid(False)
        self.verticalHeader().hide()
        self.setSelectionMode(self.SingleSelection)
        self.setSelectionBehavior(self.SelectRows)
        self.setVerticalScrollMode(self.ScrollPerPixel)
        self.setHorizontalScrollMode(self.ScrollPerPixel)
        self._stretch = 0
        self._previous_sort = 0

    def setStretch(self, column):
        self._stretch = column

    def refresh(self):
        header = self.horizontalHeader()
        QtCompat.setSectionResizeMode(
            header, self._stretch, QtWidgets.QHeaderView.Stretch)

    def setModel(self, model_):
        model_.rowsInserted.connect(self.refresh)
        model_.modelReset.connect(self.refresh)
        super(SlimTableView, self).setModel(model_)
        self.refresh()

        if isinstance(model_, model.ProxyModel):
            self.doSort.connect(model_.doSort)
            model_.askOrder.connect(self.setSorting)
            self.setSortingEnabled(True)

            # Start out unsorted
            header = self.horizontalHeader()
            header.setSortIndicatorShown(False)
            header.setSortIndicator(0, QtCore.Qt.DescendingOrder)
            self.doSort.emit(-1, QtCore.Qt.DescendingOrder)

    def setSorting(self, column, order):
        header = self.horizontalHeader()
        is_sorted = header.isSortIndicatorShown()
        is_previous = self._previous_sort == column
        is_ascending = header.sortIndicatorOrder() == QtCore.Qt.AscendingOrder

        if is_ascending and is_sorted and is_previous:
            header.setSortIndicator(column, QtCore.Qt.DescendingOrder)
            header.setSortIndicatorShown(False)
            column = -1
        else:
            header.setSortIndicatorShown(True)

        self._previous_sort = column
        self.doSort.emit(column, order)

    def mousePressEvent(self, event):
        """Do request a context menu, but on mouse *press*, not release"""

        # Important, call this first to have the item be selected,
        # as per default, and *then* ask for a context menu. That
        # way, the menu and selection aligns.
        try:
            return super(SlimTableView, self).mousePressEvent(event)

        finally:
            if event.button() == QtCore.Qt.RightButton:
                self.customContextMenuRequested.emit(event.pos())


class PushButtonWithMenu(QtWidgets.QPushButton):
    def __init__(self, *args, **kwargs):
        super(PushButtonWithMenu, self).__init__(*args, **kwargs)
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)

    def mousePressEvent(self, event):
        """Do request a context menu, but on mouse *press*, not release"""

        # Important, call this first to have the item be selected,
        # as per default, and *then* ask for a context menu. That
        # way, the menu and selection aligns.
        try:
            return super(PushButtonWithMenu, self).mousePressEvent(event)

        finally:
            if event.button() == QtCore.Qt.RightButton:
                self.customContextMenuRequested.emit(event.pos())


class LineEditWithCompleter(QtWidgets.QLineEdit):
    changed = QtCore.Signal(str)

    def __init__(self, parent=None):
        super(LineEditWithCompleter, self).__init__(parent)

        proxy = QtCore.QSortFilterProxyModel(self)
        proxy.setFilterCaseSensitivity(QtCore.Qt.CaseInsensitive)

        completer = QtWidgets.QCompleter(proxy, self)
        completer.setCaseSensitivity(QtCore.Qt.CaseInsensitive)
        completer.setCompletionMode(completer.UnfilteredPopupCompletion)
        completer.setMaxVisibleItems(15)

        self.setCompleter(completer)
        self.editingFinished.connect(self.onEditingFinished)

        self._completer = completer
        self._focused = False
        self._proxy = proxy
        self._current = None

    def setModel(self, model):
        self._proxy.setSourceModel(model)
        self._completer.setModel(self._proxy)

    def setText(self, text):

        # Keep track of a "default" such that we can revert back to
        # it following a bad completion.
        self._current = text

        return super(LineEditWithCompleter, self).setText(text)

    def resetText(self):
        self.setText(self._current)

    def mousePressEvent(self, event):
        super(LineEditWithCompleter, self).mousePressEvent(event)

        # Automatically show dropdown on select
        self._completer.complete()
        self.selectAll()

    def onEditingFinished(self):
        # For whatever reason, the completion prefix isn't updated
        # when coming from the user selecting an item from the
        # completion listview. (Windows 10, PySide2, Python 3.7)
        self._completer.setCompletionPrefix(self.text())

        suggested = self._completer.currentCompletion()

        if not suggested:
            return self.resetText()

        self.changed.emit(suggested)


class MenuWithTooltip(QtWidgets.QMenu):
    """QMenu typically doesn't draw tooltips"""

    def event(self, event):
        if event.type() == QtCore.QEvent.ToolTip and self.activeAction() != 0:

            try:
                tooltip = self.activeAction().toolTip()
            except AttributeError:
                # Can sometimes return None
                pass
            else:
                QtWidgets.QToolTip.showText(
                    event.globalPos(),
                    tooltip
                )

        # Account for QStatusBar queries about tooltip
        # These typically are only available on hovering
        # an action for some time, which isn't reflected
        # in querying the menu itself for a tooltip.
        elif event.type() == QtCore.QEvent.MouseMove:
            try:
                tooltip = self.activeAction().toolTip()

                # Some tooltips are multi-line, and the statusbar
                # typically ignores newlines and writes it all out
                # as one long line.
                tooltip = tooltip.splitlines()[0]

                self.setToolTip(tooltip)

            except (AttributeError, IndexError):
                pass

            QtWidgets.QToolTip.hideText()

        return super(MenuWithTooltip, self).event(event)


class CssHighlighter(QtGui.QSyntaxHighlighter):
    def __init__(self, parent=None):
        super(CssHighlighter, self).__init__(parent)

        keyword_format = QtGui.QTextCharFormat()
        keyword_format.setForeground(QtCore.Qt.darkBlue)
        keyword_format.setFontWeight(QtGui.QFont.Bold)

        keywordPatterns = [
            "\\bchar\\b", "\\bclass\\b", "\\bconst\\b",
            "\\bdouble\\b", "\\benum\\b", "\\bexplicit\\b", "\\bfriend\\b",
            "\\binline\\b", "\\bint\\b", "\\blong\\b", "\\bnamespace\\b",
            "\\boperator\\b", "\\bprivate\\b", "\\bprotected\\b",
            "\\bpublic\\b", "\\bshort\\b", "\\bsignals\\b", "\\bsigned\\b",
            "\\bslots\\b", "\\bstatic\\b", "\\bstruct\\b",
            "\\btemplate\\b", "\\btypedef\\b", "\\btypename\\b",
            "\\bunion\\b", "\\bunsigned\\b", "\\bvirtual\\b", "\\bvoid\\b",
            "\\bvolatile\\b"
        ]

        self.rules = [
            (QtCore.QRegExp(pattern), keyword_format)
            for pattern in keywordPatterns
        ]

        class_format = QtGui.QTextCharFormat()
        class_format.setFontWeight(QtGui.QFont.Bold)
        class_format.setForeground(QtCore.Qt.darkMagenta)
        self.rules.append((
            QtCore.QRegExp("\\bQ[A-Za-z]+\\b"),
            class_format
        ))

        id_format = QtGui.QTextCharFormat()
        id_format.setFontWeight(QtGui.QFont.Bold)
        id_format.setForeground(QtCore.Qt.darkBlue)
        self.rules.append((
            QtCore.QRegExp(r"#\w{1,}\b"),
            id_format
        ))

        comment_format = QtGui.QTextCharFormat()
        comment_format.setForeground(QtCore.Qt.red)
        self.rules.append((
            QtCore.QRegExp("//[^\n]*"),
            comment_format
        ))

        self.multicomment_format = QtGui.QTextCharFormat()
        self.multicomment_format.setForeground(QtCore.Qt.gray)

        quote_format = QtGui.QTextCharFormat()
        quote_format.setForeground(QtCore.Qt.darkGreen)
        self.rules.append(
            (QtCore.QRegExp("\".*\""), quote_format)
        )

        function_format = QtGui.QTextCharFormat()
        function_format.setFontItalic(True)
        function_format.setForeground(QtCore.Qt.blue)
        self.rules.append((
            QtCore.QRegExp("\\b[A-Za-z0-9_]+(?=\\()"),
            function_format
        ))

        self.comment_start_exp = QtCore.QRegExp("/\\*")
        self.comment_end_exp = QtCore.QRegExp("\\*/")

    def highlightBlock(self, text):
        for pattern, format in self.rules:
            expression = QtCore.QRegExp(pattern)
            index = expression.indexIn(text)
            while index >= 0:
                length = expression.matchedLength()
                self.setFormat(index, length, format)
                index = expression.indexIn(text, index + length)

        self.setCurrentBlockState(0)

        startIndex = 0
        if self.previousBlockState() != 1:
            startIndex = self.comment_start_exp.indexIn(text)

        while startIndex >= 0:
            endIndex = self.comment_end_exp.indexIn(text, startIndex)

            if endIndex == -1:
                self.setCurrentBlockState(1)
                commentLength = len(text) - startIndex
            else:
                commentLength = (
                    endIndex -
                    startIndex +
                    self.comment_end_exp.matchedLength()
                )

            self.setFormat(
                startIndex,
                commentLength,
                self.multicomment_format
            )
            startIndex = self.comment_start_exp.indexIn(
                text, startIndex + commentLength)



"""The view may access a controller, but not vice versa"""

import os
import logging

from itertools import chain
from functools import partial
from collections import OrderedDict as odict

from .vendor.Qt import QtWidgets, QtCore, QtGui
from .vendor import qargparse
from .version import version
from . import resources as res, dock, model
from . import allzparkconfig

px = res.px


class Window(QtWidgets.QMainWindow):
    title = "Allzpark %s" % version

    def __init__(self, ctrl, parent=None):
        super(Window, self).__init__(parent)
        self.setWindowTitle(self.title)
        self.setAttribute(QtCore.Qt.WA_StyledBackground)
        self.setWindowIcon(QtGui.QIcon(res.find("Logo_64.png")))

        self._count = 0

        pages = odict((
            ("home", QtWidgets.QWidget()),

            # Pages matching a particular state
            ("booting", QtWidgets.QWidget()),
            ("loading", QtWidgets.QWidget()),
            ("errored", QtWidgets.QWidget()),
            ("noapps", QtWidgets.QWidget()),
            ("noprofiles", QtWidgets.QWidget()),
        ))

        panels = {
            "pages": QtWidgets.QStackedWidget(),
            "header": QtWidgets.QWidget(),
            "body": QtWidgets.QWidget(),
            "footer": QtWidgets.QWidget(),
        }

        widgets = {
            "bootMessage": QtWidgets.QLabel("Loading.."),
            "loadingMessage": QtWidgets.QLabel("Loading"),
            "errorMessage": QtWidgets.QLabel("Uh oh..<br>"
                                             "See Console for details"),
            "noappsMessage": QtWidgets.QTextBrowser(),
            "noprofileMessage": QtWidgets.QLabel("No profiles was found"),
            "pkgnotfoundMessage": QtWidgets.QLabel(
                "One or more packages could not be found"
            ),

            # Header
            "logo": QtWidgets.QToolButton(),
            "appVersion": QtWidgets.QLabel(version),

            "profileBtn": dock.PushButtonWithMenu(),
            "profileName": dock.LineEditWithCompleter(),
            "profileVersion": dock.LineEditWithCompleter(),

            "apps": dock.SlimTableView(),
            "fullCommand": FullCommand(ctrl),

            # Error page
            "continue": QtWidgets.QPushButton("Continue"),
            "reset": QtWidgets.QPushButton("Reset"),

            "dockToggles": QtWidgets.QWidget(),

            "stateIndicator": QtWidgets.QLabel(),
        }

        # The order is reflected in the UI
        docks = odict((
            ("app", dock.App(ctrl)),
            ("packages", dock.Packages(ctrl)),
            ("context", dock.Context(ctrl)),
            ("environment", dock.Environment(ctrl)),
            ("console", dock.Console()),
            ("commands", dock.Commands()),
            ("preferences", dock.Preferences(self, ctrl)),
        ))

        # Expose to CSS
        for name, widget in chain(panels.items(),
                                  widgets.items(),
                                  pages.items()):
            widget.setAttribute(QtCore.Qt.WA_StyledBackground)
            widget.setObjectName(name)

        self.setCentralWidget(panels["pages"])

        # Add header to top-most portion of GUI above movable docks
        toolbar = self.addToolBar("header")
        toolbar.setObjectName("Header")
        toolbar.addWidget(panels["header"])
        toolbar.setMovable(False)

        # Fill horizontal space
        panels["header"].setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                       QtWidgets.QSizePolicy.Preferred)

        for page in pages.values():
            panels["pages"].addWidget(page)

        # Layout
        layout = QtWidgets.QGridLayout(pages["home"])
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(panels["body"], 1, 0)

        layout = QtWidgets.QVBoxLayout(pages["booting"])
        layout.addWidget(QtWidgets.QWidget(), 1)
        layout.addWidget(widgets["bootMessage"], 0, QtCore.Qt.AlignHCenter)
        layout.addWidget(QtWidgets.QWidget(), 1)

        layout = QtWidgets.QVBoxLayout(pages["loading"])
        layout.addWidget(QtWidgets.QWidget(), 1)
        layout.addWidget(widgets["loadingMessage"], 0, QtCore.Qt.AlignHCenter)
        layout.addWidget(QtWidgets.QWidget(), 1)

        layout = QtWidgets.QVBoxLayout(pages["errored"])
        layout.addWidget(QtWidgets.QWidget(), 1)
        layout.addWidget(widgets["errorMessage"], 0, QtCore.Qt.AlignHCenter)
        layout.addWidget(widgets["continue"], 0, QtCore.Qt.AlignHCenter)
        layout.addWidget(widgets["reset"], 0, QtCore.Qt.AlignHCenter)
        layout.addWidget(QtWidgets.QWidget(), 1)

        layout = QtWidgets.QVBoxLayout(pages["noapps"])
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(widgets["noappsMessage"], 0)

        layout = QtWidgets.QVBoxLayout(pages["noprofiles"])
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(QtWidgets.QWidget(), 1)
        layout.addWidget(widgets["noprofileMessage"], 0, QtCore.Qt.AlignHCenter)
        layout.addWidget(QtWidgets.QWidget(), 1)

        #  _______________________________________________________
        # |          |         |         |               |        |
        # |   logo   | profile |---------|               |--------|
        # |__________|_________|_________|_______________|________|
        #

        layout = QtWidgets.QGridLayout(panels["header"])
        layout.setHorizontalSpacing(px(10))
        layout.setVerticalSpacing(0)

        def addColumn(widgets, *args, **kwargs):
            """Convenience function for adding columns to GridLayout"""
            addColumn.row = getattr(addColumn, "row", -1)
            addColumn.row += kwargs.get("offset", 1)

            for row, widget in enumerate(widgets):
                layout.addWidget(widget, row, addColumn.row, *args)

            if kwargs.get("stretch"):
                layout.setColumnStretch(addColumn.row, kwargs["stretch"])

        addColumn([widgets["profileBtn"]], 2, 1)
        addColumn([widgets["profileName"],
                   widgets["profileVersion"]])

        addColumn([QtWidgets.QWidget()], stretch=1)
        addColumn([widgets["dockToggles"]], 2, 1)
        addColumn([QtWidgets.QWidget()], stretch=2)

        addColumn([QtWidgets.QLabel("allzpark"),
                   widgets["appVersion"]])
        addColumn([widgets["logo"]], 2, 1)

        layout = QtWidgets.QHBoxLayout(widgets["dockToggles"])
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        for name, widget in docks.items():
            toggle = QtWidgets.QPushButton()
            toggle.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                 QtWidgets.QSizePolicy.Expanding)
            toggle.setObjectName(name + "Toggle")
            toggle.setCheckable(True)
            toggle.setFlat(True)
            toggle.setProperty("type", "toggle")
            toggle.setIcon(res.icon(widget.icon))
            toggle.setIconSize(QtCore.QSize(px(32), px(32)))
            toggle.setToolTip("\n".join([
                type(widget).__name__, widget.__doc__ or ""]))

            def on_toggled(widget, toggle):
                widget.setVisible(toggle.isChecked())
                self.on_dock_toggled(widget, toggle.isChecked())

            def on_visible(widget, toggle, state):
                toggle.setChecked(widget.isVisible())

            toggle.clicked.connect(partial(on_toggled, widget, toggle))

            # Store reference for showEvent
            widget.toggle = toggle

            # Store reference for update_advanced_controls
            toggle.dock = widget

            # Create two-way connection; when widget is programatically
            # closed, or closed by other means, update toggle to reflect this.
            widget.visibilityChanged.connect(
                partial(on_visible, widget, toggle))

            # Forward any messages
            widget.message.connect(self.tell)

            layout.addWidget(toggle)

        layout = QtWidgets.QVBoxLayout(panels["body"])
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(widgets["apps"])
        layout.addWidget(widgets["fullCommand"])

        status_bar = self.statusBar()
        status_bar.addPermanentWidget(widgets["stateIndicator"])

        # Setup
        css = "QWidget { border-image: url(%s); }"
        widgets["logo"].setStyleSheet(css % res.find("Logo_64"))

        widgets["noappsMessage"].setReadOnly(True)
        widgets["noappsMessage"].setOpenExternalLinks(True)

        widgets["profileBtn"].setIconSize(QtCore.QSize(px(32), px(32)))
        widgets["profileBtn"].setToolTip("Click to change profile\n"
                                         "Right-click for options")
        widgets["profileBtn"].setIcon(res.icon("Default_Profile"))

        css = "QWidget { border-image: url(%s); }"

        widgets["logo"].setCursor(QtCore.Qt.PointingHandCursor)
        widgets["logo"].setToolTip(allzparkconfig.help_url)
        widgets["profileBtn"].setCursor(QtCore.Qt.PointingHandCursor)
        widgets["profileName"].setCursor(QtCore.Qt.PointingHandCursor)
        widgets["profileVersion"].setCursor(QtCore.Qt.PointingHandCursor)

        widgets["profileName"].setToolTip("Click to change profile")
        widgets["profileName"].setModel(ctrl.models["profileNames"])
        widgets["profileVersion"].setToolTip("Click to change profile version")
        widgets["profileVersion"].setModel(ctrl.models["profileVersions"])

        docks["packages"].set_model(ctrl.models["packages"])
        docks["context"].set_model(ctrl.models["context"])
        docks["environment"].set_model(ctrl.models["environment"])
        docks["commands"].set_model(ctrl.models["commands"])

        proxy_model = model.ProxyModel(ctrl.models["apps"])
        widgets["apps"].setModel(proxy_model)

        widgets["errorMessage"].setAlignment(QtCore.Qt.AlignHCenter)

        widgets["logo"].clicked.connect(self.on_logo_clicked)
        widgets["profileBtn"].clicked.connect(self.on_profilebtn_pressed)
        widgets["profileBtn"].customContextMenuRequested.connect(
            self.on_profilebtn_contextmenu)
        widgets["profileName"].changed.connect(self.on_profilename_changed)
        widgets["profileVersion"].changed.connect(
            self.on_profileversion_changed)

        widgets["reset"].clicked.connect(self.on_reset_clicked)
        widgets["continue"].clicked.connect(self.on_continue_clicked)
        widgets["apps"].activated.connect(self.on_app_clicked)
        widgets["apps"].customContextMenuRequested.connect(
            self.on_app_right_click)

        selection_model = widgets["apps"].selectionModel()
        selection_model.selectionChanged.connect(self.on_app_selection_changed)

        ctrl.models["apps"].modelReset.connect(self.on_apps_reset)
        ctrl.models["profileNames"].modelReset.connect(
            self.on_profilename_reset)
        ctrl.models["profileVersions"].modelReset.connect(
            self.on_profileversion_reset)
        ctrl.resetted.connect(self.on_reset)
        ctrl.state_changed.connect(self.on_state_changed)
        ctrl.logged.connect(self.on_logged)
        ctrl.profile_changed.connect(self.on_profile_changed)
        ctrl.repository_changed.connect(self.on_repository_changed)
        ctrl.command_changed.connect(self.on_command_changed)
        ctrl.application_changed.connect(self.on_app_changed)

        self._pages = pages
        self._widgets = widgets
        self._panels = panels
        self._docks = docks
        self._ctrl = ctrl

        self.setup_docks()
        self.on_state_changed("booting")
        self.update_advanced_controls()
        self.setFocus()

        # Enable mouse tracking for tooltips
        QtWidgets.QApplication.instance().installEventFilter(self)

    def eventFilter(self, obj, event):
        """Forward tooltips to status bar whenever the mouse moves"""
        if event.type() == QtCore.QEvent.MouseMove:
            try:
                tooltip = obj.toolTip()

                # Some tooltips are multi-line, and the statusbar
                # typically ignores newlines and writes it all out
                # as one long line.
                tooltip = tooltip.splitlines()[0]

                self.statusBar().showMessage(tooltip, 2000)
            except (AttributeError, IndexError):
                self.statusBar().clearMessage()

        # Forward the event to subsequent listeners
        return False

    def createPopupMenu(self):
        """Disables default popup for QDockWidget and QToolBar"""

    def update_advanced_controls(self):
        shown = bool(self._ctrl.state.retrieve("showAdvancedControls"))
        self._widgets["profileVersion"].setVisible(shown)
        self._widgets["fullCommand"].setVisible(shown)

        # Update dock toggles
        toggles = self._widgets["dockToggles"].layout()
        for index in range(toggles.count()):
            item = toggles.itemAt(index)
            widget = item.widget()
            dock = widget.dock

            visible = (not dock.advanced) or shown
            widget.setVisible(visible)

            if not visible:
                dock.hide()

    def setup_docks(self):
        for widget in self._docks.values():
            widget.hide()

        self.setTabPosition(QtCore.Qt.RightDockWidgetArea,
                            QtWidgets.QTabWidget.North)
        self.setTabPosition(QtCore.Qt.LeftDockWidgetArea,
                            QtWidgets.QTabWidget.North)

        area = QtCore.Qt.RightDockWidgetArea
        first = list(self._docks.values())[0]
        self.addDockWidget(area, first)

        for widget in self._docks.values():
            if widget is first:
                continue

            self.addDockWidget(area, widget)
            self.tabifyDockWidget(first, widget)

    def reset(self):
        self._ctrl.reset()

    def on_reset(self):
        pass

    def on_command_changed(self, command):
        self._widgets["fullCommand"].setText(command)

    def on_command_copied(self, cmd):
        self.tell("Copied command '%s'" % cmd)

    def on_reset_clicked(self):
        self.reset()

    def on_continue_clicked(self):
        self._ctrl.state.to_ready()

    def on_app_right_click(self, position):
        view = self._widgets["apps"]
        index = view.indexAt(position)

        if not index.isValid():
            # Clicked outside any item
            return

        model = index.model()
        tools = model.data(index, "tools")

        menu = dock.MenuWithTooltip(self)

        separator = QtWidgets.QWidgetAction(menu)
        separator.setDefaultWidget(QtWidgets.QLabel("Quick Launch"))
        menu.addAction(separator)

        def handle(action):
            tool = action.text()
            self._ctrl.launch(command=tool)

        for tool in tools:
            action = QtWidgets.QAction(tool, menu)
            action.triggered.connect(lambda _=False, a=action: handle(a))
            menu.addAction(action)

        menu.addSeparator()

        extras = (
            ["start cmd", "start powershell"]
            if os.name == "nt" else
            ["bash"]
        )

        for tool in extras:
            action = QtWidgets.QAction(tool, menu)
            action.triggered.connect(lambda _=False, a=action: handle(a))
            menu.addAction(action)

        menu.move(QtGui.QCursor.pos())
        menu.show()

    def on_setting_changed(self, argument):
        if isinstance(argument, qargparse.Button):
            if argument["name"] == "resetLayout":
                self.tell("Restoring layout..")
                geometry = self._ctrl.state.retrieve("default/geometry")
                window = self._ctrl.state.retrieve("default/windowState")
                self.restoreGeometry(geometry)
                self.restoreState(window)
            return

        key = argument["name"]
        value = argument.read()

        self.tell("Storing %s = %s" % (key, value))
        self._ctrl.state.store(argument["name"], argument.read())

        # Subsequent settings are stored to disk
        if key == "showAdvancedControls":
            self.update_advanced_controls()

        if key in ("showAllApps",
                   "showHiddenApps",
                   "patchWithFilter"):
            self._ctrl.reset()

        if key == "exclusionFilter":
            allzparkconfig.exclude_filter = value
            self._ctrl.reset()

    def on_dock_toggled(self, dock, visible):
        """Make toggled dock the active dock"""

        if not visible:
            return

        # Handle the easy cases first
        app = QtWidgets.QApplication.instance()
        ctrl_held = app.keyboardModifiers() & QtCore.Qt.ControlModifier
        allow_multiple = bool(self._ctrl.state.retrieve("allowMultipleDocks"))

        if ctrl_held or not allow_multiple:
            for d in self._docks.values():
                d.setVisible(d == dock)
            return

        # Otherwise we'll want to make the newly visible dock the active tab.

        # Turns out to not be that easy
        # https://forum.qt.io/topic/42044/
        # tabbed-qdockwidgets-how-to-fetch-the-qwidgets-under-a-qtabbar/10

        # TabBar's are dynamically created as the user
        # moves docks around, and not all of them are
        # visible or in use at all times. (Poor garbage collection)
        bars = self.findChildren(QtWidgets.QTabBar)

        # The children of a QTabBar isn't the dock directly, but rather
        # the buttons in the tab, which are of type QToolButton.

        uid = dock.windowTitle()  # note: This must be unique

        # Find which tab is associated to this QDockWidget, if any
        def find_dock(bar):
            for index in range(bar.count()):
                if uid == bar.tabText(index):
                    return index

        for bar in bars:
            index = find_dock(bar)

            if index is not None:
                break
        else:
            # Dock isn't part of any tab and is directly visible
            return

        bar.setCurrentIndex(index)

    def on_profilename_clicked(self):
        pass

    def on_profilename_changed(self, profile):
        """User changed the profile"""
        self._ctrl.select_profile(profile)
        self._ctrl.state.store("startupProfile", profile)
        self.setFocus()

    def on_profileversion_changed(self, version):
        """User changed the profile version"""
        profile = self._ctrl.current_profile
        self._ctrl.select_profile(profile, version)
        self.setFocus()

    def on_profilebtn_pressed(self):
        widget = self._widgets["profileName"]
        widget.setFocus()
        widget.selectAll()

        completer = widget.completer()
        completer.complete()

    def on_profilebtn_contextmenu(self):
        name = self._widgets["profileName"].text()

        menu = dock.MenuWithTooltip(self)
        separator = QtWidgets.QWidgetAction(menu)
        separator.setDefaultWidget(QtWidgets.QLabel(name))
        menu.addAction(separator)

        def on_reset():
            self.reset()

        reset = QtWidgets.QAction("Reset", menu)
        reset.triggered.connect(on_reset)
        reset.setToolTip("Re-scan repository for new Rez packages")
        menu.addAction(reset)

        menu.addSeparator()

        menu.move(QtGui.QCursor.pos())
        menu.show()

    def on_logo_clicked(self):
        url = allzparkconfig.help_url
        QtGui.QDesktopServices.openUrl(QtCore.QUrl(url))

    def on_profile_changed(self, profile, version, refreshed=False):
        """Profile changed in the controller"""

        # Happens when editing requirements
        action = "Refreshing" if refreshed else "Changing"
        icon = res.find("Default_Profile")
        label = profile

        package = self._ctrl.state["rezProfiles"][profile][version]
        data = allzparkconfig.metadata_from_package(package)
        label = data["label"]

        # Facilitate overriding of icon via package metadata
        if data.get("icon"):
            try:
                values = {
                    "root": os.path.dirname(package.uri),
                    "width": px(32),
                    "height": px(32),
                }

                icon = data["icon"].format(**values).replace("\\", "/")

            except KeyError:
                self.tell("Misformatted %s.icon" % package.name)

            except TypeError:
                self.tell("Unsupported package repository "
                          "for icon of %s" % package.uri)

            except Exception:
                self.tell("Unexpected error coming from icon of %s"
                          % package.uri)

        self.tell("%s %s-%s" % (action, profile, version))
        self.setWindowTitle("%s  |  %s" % (label, self.title))

        self._widgets["profileName"].setText(label)
        self._widgets["profileVersion"].setText(version)

        button = self._widgets["profileBtn"]
        button.setIcon(res.icon(icon))

        # Determine aspect ratio
        height = px(32)
        pixmap = res.pixmap(icon)

        if pixmap.isNull():
            pixmap = res.pixmap("Default_Profile")

        pixmap = pixmap.scaledToHeight(height)
        width = pixmap.width()

        button.setIconSize(QtCore.QSize(width, height))
        button.setAutoFillBackground(True)

    def setStyleSheet(self, style):
        style = style % {
            "root": os.path.dirname(__file__).replace("\\", "/"),
            "res": res.dirname.replace("\\", "/"),
        }
        super(Window, self).setStyleSheet(style)

    def on_repository_changed(self):
        self.reset()

    def on_show_error(self):
        self._docks["console"].append(self._ctrl.current_error)
        self._docks["console"].raise_()

    def tell(self, message, level=logging.INFO):
        self._docks["console"].append(message, level)
        self.statusBar().showMessage(message, 2000)

    def on_logged(self, message, level):
        self._docks["console"].append(message, level)

    def on_state_changed(self, state):
        self.tell("State: %s" % state, logging.DEBUG)

        page = self._pages.get(str(state), self._pages["home"])
        page_name = page.objectName()
        self._panels["pages"].setCurrentWidget(page)
        self._widgets["profileName"].setEnabled(True)
        self._widgets["profileVersion"].setEnabled(True)
        self._widgets["profileBtn"].setEnabled(True)

        launch_btn = self._docks["app"]._widgets["launchBtn"]
        launch_btn.setText("Launch")
        launch_btn.setEnabled(True)

        for widget in self._docks.values():
            widget.setEnabled(True)

        if page_name == "home":
            self._widgets["apps"].setEnabled(state == "ready")
            self._widgets["profileBtn"].setEnabled(state == "ready")
            self._widgets["profileVersion"].setEnabled(state == "ready")

        elif page_name == "noapps":
            message = self._ctrl.state["error"]
            self._widgets["profileBtn"].setEnabled(True)
            self._widgets["noappsMessage"].setText(message)

        elif page_name == "noprofiles":
            self._widgets["profileBtn"].setEnabled(True)

        if state == "launching":
            self._docks["app"].setEnabled(False)

        if state == "loading":
            self._widgets["profileBtn"].setEnabled(False)
            self._widgets["profileName"].setEnabled(False)
            self._widgets["profileVersion"].setEnabled(False)
            for widget in self._docks.values():
                widget.setEnabled(False)

        if state in ("pkgnotfound", "errored"):
            console = self._docks["console"]
            console.show()
            self.on_dock_toggled(console, visible=True)

            page = self._pages["errored"]
            self._panels["pages"].setCurrentWidget(page)

            self._widgets["apps"].setEnabled(False)
            launch_btn.setEnabled(False)
            launch_btn.setText("Package not found")

        if state == "notresolved":
            self._widgets["apps"].setEnabled(False)
            launch_btn.setEnabled(False)
            launch_btn.setText("Failed to resolve")

        self._widgets["stateIndicator"].setText(str(state))
        self.update_advanced_controls()

    def on_launch_clicked(self):
        self._ctrl.launch()

    def on_profilename_reset(self):
        pass

    def on_profileversion_reset(self):
        pass

    def on_apps_reset(self):
        app = self._ctrl.state.retrieve("startupApplication")

        row = 0
        model = self._ctrl.models["apps"]

        if app:
            for row_ in range(model.rowCount()):
                index = model.index(row_, 0, QtCore.QModelIndex())
                name = model.data(index, "name")

                if app == name:
                    row = row_
                    self.tell("Using startup application %s" % name)
                    break

        self._widgets["apps"].selectRow(row)

    def on_app_clicked(self, index):
        """An app was double-clicked or Return was hit"""

        app = self._docks["app"]
        app.show()
        self.on_dock_toggled(app, visible=True)

    def on_app_selection_changed(self, selected, deselected):
        """The current app was changed

        Arguments:
            selected (QtCore.QItemSelection): ..
            deselected (QtCore.QItemSelection): ..

        """

        try:
            index = selected.indexes()[0]
        except IndexError:
            # No app was selected
            return

        model = index.model()
        app_request = model.data(index, "name")
        self._ctrl.select_application(app_request)

    def on_app_changed(self):
        selection_model = self._widgets["apps"].selectionModel()
        index = selection_model.selectedIndexes()[0]
        self._docks["app"].refresh(index)

    def showEvent(self, event):
        super(Window, self).showEvent(event)
        self._ctrl.state.store("default/geometry", self.saveGeometry())
        self._ctrl.state.store("default/windowState", self.saveState())

        if self._ctrl.state.retrieve("geometry"):
            self.tell("Restoring layout..", logging.DEBUG)
            self.restoreGeometry(self._ctrl.state.retrieve("geometry"))
            self.restoreState(self._ctrl.state.retrieve("windowState"))

    def closeEvent(self, event):
        self._ctrl.state.store("geometry", self.saveGeometry())
        self._ctrl.state.store("windowState", self.saveState())
        return super(Window, self).closeEvent(event)


class FullCommand(QtWidgets.QWidget):
    def __init__(self, ctrl, parent=None):
        super(FullCommand, self).__init__(parent)

        widgets = {
            "options": QtWidgets.QComboBox(),
            "text": QtWidgets.QLineEdit(),
        }

        layout = QtWidgets.QHBoxLayout(self)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(widgets["options"])
        layout.addWidget(widgets["text"])

        up_icon = res.icon("Action_GoUp_3_Large_32.png")
        down_icon = res.icon("Action_GoDown_3_32.png")

        widgets["text"].setReadOnly(True)
        widgets["options"].addItem(up_icon, "Used Request")
        widgets["options"].addItem(down_icon, "Used Resolve")

        widgets["options"].currentIndexChanged.connect(self.on_option_changed)

        self._widgets = widgets
        self._ctrl = ctrl

        index = (
            0
            if ctrl.state.retrieve("serialisationMode") == "used_request"
            else 1
        )

        widgets["options"].setCurrentIndex(index)

    def setText(self, text):
        self._widgets["text"].setText(text)

    def on_option_changed(self, option):
        if option == 0:
            mode = "used_request"
            self._widgets["options"].setToolTip(
                "Used Request\n"
                "Use the final request made to Rez,\n"
                "excluding any indirect requirements."
            )
        else:
            mode = "used_resolve"
            self._widgets["options"].setToolTip(
                "Used Resolve\n"
                "Use the fully resolved list of packages,\n"
                "including packages that may not be suitable\n"
                "for another platform."
            )

        self._ctrl.update_command(mode)



# API wrapper for Rez

from rez.resolved_context import ResolvedContext as env
from rez.packages_ import iter_packages as find
from rez.package_copy import copy_package
from rez.package_filter import Rule, PackageFilterList
from rez.package_repository import package_repository_manager
from rez.packages_ import Package
from rez.utils.formatting import PackageRequest
from rez.config import config
from rez.util import which
from rez import __version__ as version
from rez.exceptions import (
    PackageFamilyNotFoundError,
    RexUndefinedVariableError,
    ResolvedContextError,
    RexError,
    PackageCommandError,
    PackageRequestError,
    PackageNotFoundError,
    RezError,
)
from rez.utils.graph_utils import save_graph


def clear_caches():
    for path in config.packages_path:
        repo = package_repository_manager.get_repository(path)
        repo.clear_caches()


def find_one(*args, **kwargs):
    return next(find(*args, **kwargs))


def find_latest(*args, **kwargs):
    return list(
        sorted(
            find(*args, **kwargs), key=lambda pkg: pkg.version
        )
    )[-1]


try:
    from rez import project
except ImportError:
    # nerdvegas/rez
    project = "rez"


__all__ = [
    "env",
    "find",
    "find_one",
    "find_latest",
    "config",
    "version",
    "project",
    "copy_package",
    "package_repository_manager",

    # Classes
    "Package",
    "PackageRequest",

    # Exceptions
    "PackageFamilyNotFoundError",
    "ResolvedContextError",
    "RexUndefinedVariableError",
    "RexError",
    "PackageCommandError",
    "PackageNotFoundError",
    "PackageRequestError",
    "RezError",

    # Filters
    "Rule",
    "PackageFilterList",

    # Extras
    "which",
    "save_graph",
    "clear_caches",
]



"""The Allzpark configuration file

Copy this onto your local drive and make modifications.
Anything not specified in your copy is inherited from here.

ALLZPARK_CONFIG_FILE=/path/to/allzparkconfig.py

"""

import os as __os


# Load this profile on startup.
# Defaults to the first available from `profiles`
startup_profile = ""  # (optional)

# Pre-select this application in the list of applications,
# if it exists in the startup profile.
startup_application = ""  # (optional)

# Default filter, editable via the Preferences page
exclude_filter = "*.beta"

# Where to go when clicking the logo
help_url = "https://allzpark.com"


def profiles():
    """Return list of profiles

    This function is called asynchronously, and is suitable
    for making complex filesystem or database queries.
    Can also be a variable of type tuple or list

    """

    try:
        return __os.listdir(__os.path.expanduser("~/profiles"))
    except IOError:
        return []


def applications():
    """Return list of applications

    Applications are typically provided by the profile,
    this function is called when "Show all apps" is enabled.

    """

    return []


def applications_from_package(variant):
    """Return applications relative `variant`

    Returns:
        list of strings: E.g. ['appA', 'appB==2019']

    """

    from . import _rezapi as rez

    # May not be defined
    requirements = variant.requires or []

    apps = list(
        str(req)
        for req in requirements
        if req.weak
    )

    # Strip the "weak" property of the request, else iter_packages
    # isn't able to find the requested versions.
    apps = [rez.PackageRequest(req.strip("~")) for req in apps]

    # Expand versions into their full range
    # E.g. maya-2018|2019 == ["maya-2018", "maya-2019"]
    flattened = list()
    for request in apps:
        flattened += rez.find(
            request.name,
            range_=request.range,
        )

    # Return strings
    apps = list(
        "%s==%s" % (package.name, package.version)
        for package in flattened
    )

    return apps


def metadata_from_package(variant):
    """Return metadata relative `variant`

    Blocking call, during change of profile.

    IMPORTANT: this function must return at least the
        members part of the original function, else the program
        will not function. Very few safeguards are put in place
        in favour of performance.

    Arguments:
        variant (rez.packages_.Variant): Package from which to retrieve data

    Returns:
        dict: See function for values and types

    """

    data = getattr(variant, "_data", {})

    return dict(data, **{

        # Guaranteed keys, with default values
        "label": data.get("label", variant.name),
        "background": data.get("background"),
        "icon": data.get("icon", ""),
        "hidden": data.get("hidden", False),
    })
